/*
    BOMB, an open-platform NUON game
	
	Programming by Andreas Binner
	
	Main module
*/

#define VERSION "V1.1"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <nuon/sprite.h>
#include <nuon/bios.h>
#include <nuon/dma.h>
#include <nuon/video.h>
#include <nuon/joystick.h>
#include <nuon/mediaio.h>
#include <nuon/nise.h>
#include <nuon/termemu.h>


#include "def.h"
#include "proto.h"
#include "levels.h"

/* Binary data */

extern long StepSound[];
extern long BombSound[];
extern long ActivateSound[];
extern long SplashSound[];
extern long IceSound[];
extern long BreakSound[];
extern long ScreamSound[];
extern long TeleportSound[];
extern long SwitchSound[];
extern long SpinachSound[];
extern long ClankSound[];

extern unsigned char StartScreenTGA[];
extern unsigned char Instr1TGA[];
extern unsigned char Instr2TGA[];
extern unsigned char LoadingTGA[];
extern unsigned char PausedTGA[];
extern unsigned char LevelCompTGA[];
extern unsigned char GameOverTGA[];

/* typedefs */

typedef struct
{
	int id;
	int x,y;
	int on_tile,off_tile;
	char on_type;
} SWITCH;

typedef struct
{
	int x,y;
} COORD;

typedef struct
{
	int x,y;
	int phase;
	int dir;
} DRONE;

typedef struct
{
	int next,prev;
	int x,y;
	int count;
	int type;
	int DropIsHit;
} BOMB;

typedef struct
{
	int ox,oy;
	int x,y;
	int type;
	unsigned int flags;
	int aux;
} TCOORD;

/* game global variables */

DRONE drones[MAX_DRONES];
int numDrones=0;

SWITCH switchtab[MAXSWITCHES];
int numSwitches=0;

BOMB bomblist[MAXBOMBS];
int bombListStart;

TCOORD tpos[32][32];
int dropX,dropY,dropDir=2,dropPhase=0,dropTimer=0,dropRotate=0,dropIdle=0;
int dropAnimation=0,dropShield=0,dropSpinach=0;
int dropStartX,dropStartY;
int bombMove;
int lastTileType;
int lastDropX;	
int lastDropY;	
int field_x=0;
int field_y=0;
int fieldDir=0;
int moveField=0;
int score=0;
int newlifeThres=2000;
int level=0;
int lives=3;
int timer=0;
int bonus=0;
int bombs=0;
int bombTime=50;
int field_width;
char *field;
char *switches;

int fps;
int vidOn=0;
int debug=0;

unsigned long *localScratch;
int localSize;

PCMPOS Pan;
void *FrameBuffer[NUM_BUFFERS];
#ifdef OVERL_DEBUG
void *OverlayBuffer;
#endif


COORD bomb_area[]={
    {-3,0},{3,0},{0,-3},{0,3},{-1,2},{-1,-2},{1,2},{1,-2},{-2,1},{-2,-1},{2,1},{2,-1},
    {-2,0},{2,0},{0,-2},{0,2},{-1,1},{1,1},{-1,-1},{1,-1},
	{-1,0},{1,0},{0,-1},{0,1},
	{0,0}
	};	


/*
 * wait for video
 * returns the number of elapsed fields
 * since the last sync
 */

int
VideoSync(int drawBuffer)
{
    static int frames;
    static int lastfield;
    static int lastsecond;
    int ret;

    int curfield;

    /* we want to view the buffer we just finished
       drawing */

    /* set the video to point at the drawn buffer */
    _VidChangeBase(VID_CHANNEL_MAIN,DMAFLAGS,FrameBuffer[drawBuffer]);

    /* find out how many fields have been drawn */
    curfield = _VidSync(-1);

    /* remember that we've drawn a frame */
    frames++;

    /* have more than 60 fields (1 second) passed? */
    if (curfield - lastsecond >= 60) {
        fps = frames;
        frames = 0;
        lastsecond = curfield;
    }

    /* draw to a new buffer */
    drawBuffer++;
    if (drawBuffer >= NUM_BUFFERS)
        drawBuffer = 0;

#if NUM_BUFFERS == 2
    /* wait until a new field starts (so it's no longer
       showing the field we're going to draw on) */
    curfield = _VidSync(0);
#endif

    ret = curfield - lastfield;
    while (ret < 3) {                /* Lock to 20 fps !!!*/
        curfield = _VidSync(-1);
        ret = curfield - lastfield;
    }
    lastfield = curfield;
    return drawBuffer;
}

/*
	Debug printf
*/

#ifndef OVERL_DEBUG
extern void kprintf(char *str, ...);
#else
void kprintf(char *str,...)
{
	char temp[256];
	va_list ap;

	if (!debug)
		return;
			
	if (strrchr(str,10))
		*strrchr(str,10)=0;
		
    va_start(ap, str);
	vsprintf(temp,str,ap);
    va_end(ap);
	if (vidOn)
		Print(temp,kWhite,0x00);
}
#endif

/*
	Add a new activated bomb to the bomb list
	
	The bomb list is a linked list with forward and backward pointers
*/

int AddBombToList(int x,int y,int type,int count)
{
	int i,free;

	free=-1;
	for(i=0;i<MAXBOMBS;i++)
		if (bomblist[i].type<0)
		{
			free=i;
			break;
		}
	
	if (free<0)
		return -1;
	
	i=bombListStart;	
	if (i>=0)
	{
		while (bomblist[i].next>=0)
			i=bomblist[i].next;
		bomblist[i].next=free;
		bomblist[free].prev=i;
	}
	else
	{
		bombListStart=free;
		bomblist[free].prev=-1;
	}
	
//	kprintf("AddBombToList %d: x:%d y:%d type:%d count:%d\n",free,x,y,type,count);
			
	bomblist[free].x=x;
	bomblist[free].y=y;
	bomblist[free].count=count;
	bomblist[free].type=type;
	bomblist[free].next=-1;

	return 0;
}

/*
	Activate a bomb at (x/y) and all bombs in its
	activation radius, 
*/

void ActivateBomb(int x,int y,int count)
{
	int i,c,bx,by;
	
	switch(tpos[x][y].type)
	{
		case SMALLMINE:
		case SMALLBOMB:
			c=SMALL_BOMB_ACT_SIZE;
		break;	
		case MEDBOMB:
			c=MEDIUM_BOMB_ACT_SIZE;
		break;
		case BIGMINE:	
		case BIGBOMB:
			c=BIG_BOMB_ACT_SIZE;
		break;
		default:
			return;	
	}

	i=AREA_START;
	while(c)
	{
		bx = bomb_area[i].x + x;
		by = bomb_area[i].y + y;
		c--;
		i--;
		if (bx<0 || by<0 || bx>=field_width || by>=field_width)
			continue;
		if (((tpos[bx][by].type>=SMALLBOMB && tpos[bx][by].type<=BIGBOMB) || tpos[bx][by].type==SMALLMINE || tpos[bx][by].type==BIGMINE)  && !(tpos[bx][by].flags&ACTIVATED))
		{
			tpos[bx][by].flags|=ACTIVATED;
			AddBombToList(bx,by,tpos[bx][by].type,count);
			ActivateBomb(bx,by,count);
		}
		if (!(c%4))
			count+=BOMB_CHAIN_DELAY;
	}

}

/*
    Wait for a joystick (or remote) button
*/

int WaitForJoystick(int but)
{
	int edge,joystick;
	int old_joystick=0;
	
    do
	{
		joystick = _Controller[1].buttons |_Controller[0].buttons;                                                                               
		edge = (joystick ^ old_joystick) & joystick; 
		old_joystick = joystick;
	} while (!(edge & but));
    do
	{
		joystick = _Controller[1].buttons |_Controller[0].buttons;                                                                               
	} while ((joystick & but));
	
	return (edge & but);
}

/*
   Convert a ASCII tile character to a type value
*/

int Char2Type(int v)
{
	int t;
	
	t=WATER;
	switch(v)
	{
		case '$':
		case '#':
			t=TILE;
		break;
		case '.':
			t=SMALLBOMB;
		break;
		case 'o':
			t=MEDBOMB;
		break;
		case 'O':
			t=BIGBOMB;
		break;
		case '*':
			t=ICE;
		break;
		case 'x':
			t=BROKEN;
		break;
		case '=':
			t=SAVETILE;
		break;
		case '<':
			t=SMALLMINE;
		break;
		case '>':
			t=BIGMINE;
		break;
		case '~':
			t=SHIELD;
		break;
		case 'S':
			t=SPINACH;
		break;
		default:
			if (v>=SWITCH_CHAR_START && v<=SWITCH_CHAR_END)
				t=SWITCH_OFF;
			if (v>=TELEPORT_CHAR_START && v<=TELEPORT_CHAR_END)
				t=TELEPORT;
		break;
	}
	return t;
}

/*
    Is given type a bomb or mine?
*/

int isBomb(int type)
{
	if ((type>=SMALLBOMB && type<=BIGBOMB) || type==SMALLMINE || type==BIGMINE)
		return 1;
	return 0;
}

/*
   Generate playfield from ASCII string definition
*/

void FieldFromASCII(int w)
{
    int x0,y0,xp,yp,x,y,i,j,sx,sy,t,v,id,width;
	char *s;
	
	bombs=0;
	numDrones=0;
/* Build switch array */
	i=0;
	s=switches;
	while(*s)
	{
		id=(int)(*s++-'a');
		switchtab[id].on_type=*s;	
		switchtab[id].on_tile=Char2Type(*s++);	
		switchtab[id].off_tile=Char2Type(*s++);
		if (isBomb(switchtab[id].on_tile) && !isBomb(switchtab[id].off_tile))
			bombs++;
		i++;	
	}
	numSwitches=i;
/* Create field */

	width=w;
	if (w&1) w++;
    x0=-(w/2);
    y0=w/2-1;
    xp=0;
	yp=0;
    i=j=0;
    x=x0;
    y=y0;
    t=0;
    
    sx=x;
    sy=y;
    for(;;)
    {		
        if (x<0 || x>=width || y<0 || y>=width)
            v=' ';
        else
        {
		    v=field[x+field_width*y];
			tpos[x][y].ox=tpos[x][y].x=xp;
			tpos[x][y].oy=tpos[x][y].y=yp;
        }    

		if (v!=' ')
		{
			if (v=='?' && numDrones<MAX_DRONES-1)
			{
				drones[numDrones].dir=0;
				drones[numDrones].phase=0;
				drones[numDrones].x=x;
				drones[numDrones++].y=y;
				v='#';
			}

			tpos[x][y].type=Char2Type(v);
			if (v=='$')
			{
				dropX=dropStartX=x;
				dropY=dropStartY=y;
			}
			
			if (v>=TELEPORT_CHAR_START && v<=TELEPORT_CHAR_END)
				tpos[x][y].aux=v-TELEPORT_CHAR_START;

			if (v>=SWITCH_CHAR_START && v<=SWITCH_CHAR_END)
				tpos[x][y].aux=v-SWITCH_CHAR_START;
			
			if (v>='I' && v<='N')
			{
				tpos[x][y].type=TILE;
				tpos[x][y].flags|=(TELEDEST+((v-'I')<<16));
			}
				
  		    if (v>='A' && v<='F')
			{
				if (switchtab[v-'A'].on_type>=SWITCH_CHAR_START && switchtab[v-'A'].on_type<=SWITCH_CHAR_END)
					tpos[x][y].aux=switchtab[v-'A'].on_type-SWITCH_CHAR_START;
				switchtab[v-'A'].x=x;
				switchtab[v-'A'].y=y;
				tpos[x][y].type=switchtab[v-'A'].off_tile;
			}
			if (isBomb(tpos[x][y].type))
				bombs++;
       }
		
        x++;
        y--;
        xp+=64;
        i++;
        if (i==w)
        {
            if (!t)
                sx++;
            else
                sy++;
            
            t^=1;
            x=sx;
            y=sy;
            i=0;
            yp+=16;
            xp=t*32;
            j++;
            if (j==w*2)
                break;
        }
    } 

	kprintf("Bombs=%d\n",bombs);
}

/*
    Draw playfield
*/

void RenderField(int b,int w,int offx,int offy)
{
	int x,y;
	
	
    for(y=0;y<field_width;y++)
	    for(x=0;x<field_width;x++)
		{
			tpos[x][y].x=tpos[x][y].ox+offx;
			tpos[x][y].y=tpos[x][y].oy+offy;
			if (tpos[x][y].type!=WATER)
				TileCopy(b,tpos[x][y].type,tpos[x][y].x,tpos[x][y].y,40,1);
		}
}

/*
   Draw background
*/

void RenderWater(int b)
{
	int x,y;
	
	for(y=0;y<240;y+=40)
		for(x=0;x<384;x+=64)
			FastTileCopy(b,WATER,x,y,40);			
		
}

/*
   Draw hero (Drop) 
*/

void RenderDrop(int b,int dir,int phase,int x,int y,int fl)
{
	TileCopy(b,DROP0_0+4*dir+phase,x+4,y-28,48,1+fl+(dropShield?4:0));
}

/*
   Reset playfield and bomb list
*/

void ResetPlayField()
{
	int x,y,i;
	
	for(y=0;y<field_width;y++)
		for(x=0;x<field_width;x++)
		{
			tpos[x][y].type=WATER;
			tpos[x][y].flags=0;
		}
	bombListStart=-1;
	for(i=0;i<MAXBOMBS;i++)
	{
		bomblist[i].DropIsHit=0;
		bomblist[i].type=-1;
		bomblist[i].next=-1;
		bomblist[i].prev=-1;
	}
}

/*
    Search teleporter destination 
*/

void Teleport(int id)
{
	int x,y;
	
	for(y=0;y<field_width;y++)
		for(x=0;x<field_width;x++)
		{
			if (x==dropX && y==dropY) continue;
			
			if (tpos[x][y].flags==TELEDEST+(id<<16)) 
			{
				dropX=x;
				dropY=y;
				return;
			}
		}
}

/*
   See if given coordinates match Drop's position (pays attention
   to its current position between tiles!)
*/

int CheckDrop(int x,int y)
{
	int dx,dy;
	
	if ((dropAnimation&0xffff)==DROP_TELEPORTING)
		return 0;

	dx=dropX;
	dy=dropY;
	if (dropPhase>=4)
		switch(dropDir)
		{
			case 0:
				dx--;
			break;	
			case 1:
				dy--;
			break;	
			case 2:
				dx++;
			break;	
			case 3:
				dy++;
			break;	
		}
	if(dx==x && dy==y)
		return 1;
	return 0;		
}

/*
   Toggle switch at (x/y), if any.
*/

int CheckSwitch(int x,int y)
{
	int dx,dy,n;
	
	if (tpos[x][y].type==SWITCH_OFF || tpos[x][y].type==SWITCH_ON)
	{
		n=tpos[x][y].aux;
				
		dx=switchtab[n].x;
		dy=switchtab[n].y;

		PlayWAV((void *)SwitchSound, -1, &Pan, 0x40000000, 0x40000000 );
		
		if (tpos[x][y].type==SWITCH_OFF)
		{
			if (tpos[dx][dy].type==switchtab[n].off_tile)
			{
				tpos[x][y].type=SWITCH_ON;
				tpos[dx][dy].type=switchtab[n].on_tile;
			}
		}
		else
		{
			if (tpos[dx][dy].type==switchtab[n].on_tile)
			{
				tpos[x][y].type=SWITCH_OFF;
				tpos[dx][dy].type=switchtab[n].off_tile;
			}
		}
		
		if (CheckDrop(dx,dy) && tpos[dx][dy].type==WATER)
			return -1;
	}
	return 0;
}

/*
  Handle drop animations
*/

int DropAnimation(int b)
{
	int x,y,cx,cy,dx,dy,p;
	
	switch(dropAnimation&0xffff)
	{
		case DROP_DROWNING:
			p=dropAnimation>>16;
			if (dropAnimation == DROP_START_DROWNING)
				PlayWAV((void *)SplashSound, -1, &Pan, 0x40000000, 0x40000000 );
			cx=cy=0;
			dx=dropX;
			dy=dropY;
			if (dx<0)
			{
				dx=0;
				cx=-32;
				cy=-20;
			}
			if (dy<0)
			{
				dy=0;
				cx=32;
				cy=-20;
			}
			if (dx>=field_width)
			{
				dx=field_width-1;
				cx=32;
				cy=20;
			}
			if (dy>=field_width)
			{
				dy=field_width-1;
				cx=-32;
				cy=20;
			}
			x=tpos[dx][dy].x;
			y=tpos[dx][dy].y;
			TileCopy(b,DROP0_0+4*dropDir+dropPhase,x+4+cx,y-28+(48-p)+cy,p,1);			
		    dropPhase=(dropPhase+1)%4;
			if (!p)
				return -1;
		break;
		
		case DROP_TELEPORTING:
			x=tpos[dropX][dropY].x;
			y=tpos[dropX][dropY].y;
			if (dropAnimation == DROP_START_TELEPORTING)
				PlayWAV((void *)TeleportSound, -1, &Pan, 0x40000000, 0x40000000 );
			if (dropAnimation == DROP_HALF_TELEPORTING)
			{
				PlayWAV((void *)TeleportSound, -1, &Pan, 0x40000000, 0x40000000 );
				Teleport(tpos[dropX][dropY].aux);
				if (tpos[dropX][dropY].type==WATER)
					dropAnimation=DROP_START_DROWNING;
			}
			RenderDrop(b,(dropDir+dropRotate)%4,2,x,y,2);
			dropRotate=(dropRotate+1)%4;
		break;
	
	}
	dropAnimation-=0x10000;
	if (dropAnimation<0)
		dropAnimation=0;
	return 0;
}


/*
   Check joystick and control Drop
*/


int ControlDrop(int b)
{
	int edge,joystick;
	static int old_joystick=0;
	int moves[]={2,3,2,1,0,1};
	int dirX[]={-4, 4, 4,-4};
	int dirY[]={-4,-4, 4, 4};
	int x,y;
	int fl,ret;
	
	joystick = _Controller[1].buttons |_Controller[0].buttons;                                                                               
	edge = (joystick ^ old_joystick) & joystick; 
	old_joystick = joystick;
	
	if (dropAnimation)
	{
		ret=DropAnimation(b);
		if (ret<0)
			return -1;
	}
	else
	{	
		dropTimer=(dropTimer+1)%1;
		if (!dropPhase)
		{
		
			if (dropSpinach && (joystick&JOY_B) && isBomb(tpos[dropX][dropY].type))
			{
				if ((joystick&JOY_DOWN)==JOY_DOWN && tpos[dropX][dropY+1].type==TILE)
				{
					bombMove=tpos[dropX][dropY].type;
					tpos[dropX][dropY].type=TILE;
					dropDir=3;
					dropPhase=1;
					dropTimer=0;
				}		
				if ((joystick&JOY_UP)==JOY_UP && tpos[dropX][dropY-1].type==TILE)
				{
					bombMove=tpos[dropX][dropY].type;
					tpos[dropX][dropY].type=TILE;
					dropDir=1;
					dropPhase=1;
					dropTimer=0;
				}
				if ((joystick&JOY_RIGHT)==JOY_RIGHT && tpos[dropX+1][dropY].type==TILE)
				{
					bombMove=tpos[dropX][dropY].type;
					tpos[dropX][dropY].type=TILE;
					dropDir=2;
					dropPhase=1;
					dropTimer=0;
				}
				if ((joystick&JOY_LEFT)==JOY_LEFT && tpos[dropX-1][dropY].type==TILE)
				{
					bombMove=tpos[dropX][dropY].type;
					tpos[dropX][dropY].type=TILE;
					dropDir=0;
					dropPhase=1;
					dropTimer=0;
				}
				if (bombMove)
					PlayWAV((void *)ClankSound, -1, &Pan, 0x40000000, 0x40000000 );
			}
			if (edge&JOY_A)
			{
				dropIdle=0;
				CheckSwitch(dropX,dropY);

				if (tpos[dropX][dropY].type>=SMALLBOMB && tpos[dropX][dropY].type<=BIGBOMB)
				{
					PlayWAV((void *)ActivateSound, -1, &Pan, 0x40000000, 0x40000000 );
					ActivateBomb(dropX,dropY,bombTime);
				}
			}
			if (!(joystick&JOY_B))
			{
				if ((joystick&JOY_DOWN)==JOY_DOWN)
				{
					dropPhase=1;
					dropDir=3;
					dropTimer=0;
				}
				if ((joystick&JOY_UP)==JOY_UP)
				{
					dropPhase=1;
					dropDir=1;
					dropTimer=0;
				}
				if ((joystick&JOY_RIGHT)==JOY_RIGHT)
				{
					dropPhase=1;
					dropDir=2;
					dropTimer=0;
				}
				if ((joystick&JOY_LEFT)==JOY_LEFT)
				{
					dropPhase=1;
					dropDir=0;
					dropTimer=0;
				}
			}
			if (dropPhase)
				dropIdle=0;
			else
				dropIdle++;
		}
		else
		{
			if (!dropTimer)
			{
				dropPhase++;
				if (dropPhase==8)
				{
					dropPhase=0;
					bonus-=5;
					if (bonus<0) bonus=0;

					switch(dropDir)
					{
						case 0:
							dropX--;
						break;	
						case 1:
							dropY--;
						break;	
						case 2:
							dropX++;
						break;	
						case 3:
							dropY++;
						break;	
					}
					if (bombMove)
						tpos[dropX][dropY].type=bombMove;
					bombMove=0;
								
					if (dropX<0 || dropX>=field_width || dropY<0 || dropY>=	field_width || tpos[dropX][dropY].type==WATER)
					{
						dropAnimation=DROP_START_DROWNING;
						return 0;
					}
					if (tpos[dropX][dropY].type==BROKEN)
						PlayWAV((void *)BreakSound, -1, &Pan, 0x40000000, 0x40000000 );
					if (tpos[dropX][dropY].type==SMALLMINE || tpos[dropX][dropY].type==BIGMINE)
					{
						PlayWAV((void *)ActivateSound, -1, &Pan, 0x40000000, 0x40000000 );
						ActivateBomb(dropX,dropY,bombTime);
					}
					if (tpos[dropX][dropY].type==TELEPORT)
						dropAnimation=DROP_START_TELEPORTING;
						
					if (tpos[dropX][dropY].type==SHIELD)
						dropShield=SHIELD_TIME;

					if (tpos[dropX][dropY].type==SPINACH)
					{
						PlayWAV((void *)SpinachSound, -1, &Pan, 0x40000000, 0x40000000 );
						dropSpinach=SPINACH_TIME;
						tpos[dropX][dropY].type=TILE;
					}
						
					if (tpos[dropX][dropY].type==ICE)
					{
						PlayWAV((void *)IceSound, -1, &Pan, 0x40000000, 0x40000000 );
						dropPhase=1;
					}

					if (lastTileType==BROKEN)
						tpos[lastDropX][lastDropY].type=WATER;

					lastTileType=tpos[dropX][dropY].type;
					lastDropX=dropX;	
					lastDropY=dropY;	
				}
			}

		}

		x=tpos[dropX][dropY].x+dirX[dropDir]*dropPhase;
		y=tpos[dropX][dropY].y+(dirY[dropDir]*dropPhase*10)/16;
		if (tpos[dropX][dropY].type==ICE || bombMove)
			dropRotate=(dropRotate+1)%4;
		else
			dropRotate=0;

		if (!dropPhase && dropDir>1 && dropIdle>50 && !dropShield && !dropSpinach)
			TileCopy(b,DROP_SLEEP_0+dropDir-2,x+4,y-28,48,1);			
		else
			RenderDrop(b,(dropDir+dropRotate)%4,moves[dropPhase%6],x,y,0);
	}
	
	if (dropX<0 || dropX>=field_width || dropY<0 || dropY>=	field_width)
		return 0;
		
#if 0
	fl=0;
	if (tpos[dropX][dropY].x<=VX+32)
		fl=1;
	if (tpos[dropX][dropY].y<=VY+20)
		fl=2;
	if (tpos[dropX][dropY].x>=VX+VW-64)
		fl=3;
	if (tpos[dropX][dropY].y>=VY+VH-40)
		fl=4;
#else
	fl=0;
	if (tpos[dropX][dropY].x-SCREEN_WIDTH/2<-32)
		fl|=1;
	if (tpos[dropX][dropY].y-SCREEN_HEIGHT/2<-32)
		fl|=2;
	if (tpos[dropX][dropY].x-SCREEN_WIDTH/2>32)
		fl|=4;
	if (tpos[dropX][dropY].y-SCREEN_HEIGHT/2>32)
		fl|=8;
#endif

	return fl;
}

/*
   "Camera" movement
*/

#if 0
void MovePlayfield(int fl)
{
	if (moveField && (!fl || fl==fieldDir))
	{
		switch(fieldDir)
		{
			case 1:	
				field_x+=FIELD_MOVE_INCR;
			break;
			case 2:
				field_y+=FIELD_MOVE_INCR;
			break;
			case 3:
				field_x-=FIELD_MOVE_INCR;
			break;
			case 4:
				field_y-=FIELD_MOVE_INCR;
			break;
		}
		moveField--;
	}
	else
		if (fl)
		{
			if (fl==1 || fl==3)
				moveField=(VW/2)/FIELD_MOVE_INCR;
			else
				moveField=(VH/2)/FIELD_MOVE_INCR;
			 fieldDir=fl;
		}
}
#else
void MovePlayfield(int fl)
{
	if (fl&1)
			field_x+=FIELD_MOVE_INCR;
	if (fl&2)
			field_y+=FIELD_MOVE_INCR;
	if (fl&4)
			field_x-=FIELD_MOVE_INCR;
	if (fl&8)
			field_y-=FIELD_MOVE_INCR;
}
#endif

/*
   Handle bombs in bomb list. Draws explosions.
*/

int HandleBombs(int b)
{
	int i=bombListStart;
	int prev,next;
	int c=0,ex,bx,by;
	int ret,j;
		
	if (i<0)
		return 0;
	
	ret=0;
	
	do
	{
		bomblist[i].count--;
		if (bomblist[i].count<=0)
		{
			if (bomblist[i].count==-7 || tpos[bomblist[i].x][bomblist[i].y].type==WATER)
			{
				if (bomblist[i].DropIsHit)
					ret=-1;
				
				score+=10;
//				bombs--;
				bomblist[i].DropIsHit=0;
				bomblist[i].type=-1;
				prev=bomblist[i].prev;
				next=bomblist[i].next;
				if (prev>=0)
				{
					bomblist[prev].next=bomblist[i].next;
					bomblist[next].prev=prev;
				}
				else
				{
					bombListStart=bomblist[i].next;
					bomblist[next].prev=-1;
				}
			}
			else
			{
				if (!bomblist[i].count)
					PlayWAV((void *)BombSound, -1, &Pan, 0x20000000, 0x08000000 );
				switch(bomblist[i].type)
				{
					case SMALLMINE:
					case SMALLBOMB:
						c=SMALL_BOMB_AREA_SIZE;
					break;	
					case MEDBOMB:
						c=MEDIUM_BOMB_AREA_SIZE;
					break;
					case BIGMINE:	
					case BIGBOMB:
						c=BIG_BOMB_AREA_SIZE;
					break;
				}
				ex=AREA_START;
				while(c)
				{
					bx = bomb_area[ex].x + bomblist[i].x;
					by = bomb_area[ex].y + bomblist[i].y;
					c--;
					ex--;
					if (bx<0 || by<0 || bx>=field_width || by>=field_width)
						continue;
					if (CheckDrop(bx,by))
					{
						bomblist[i].DropIsHit=1;
						PlayWAV((void *)ScreamSound, -1, &Pan, 0x20000000, 0x10000000 );
					}
					if (tpos[bx][by].type!=WATER && tpos[bx][by].type!=SAVETILE)
					{
						TileCopy(b,EXPLOSION_0-bomblist[i].count,tpos[bx][by].x,tpos[bx][by].y,40,1);
						if (bomblist[i].count==-6)
						{
							if (isBomb(tpos[bx][by].type))
								bombs--;
							kprintf("Bombs=%d\n",bombs);
							tpos[bx][by].type=WATER;
							for(j=0;j<numDrones;j++)
								if (drones[j].x==bx && drones[j].y==by)
									drones[j].dir=-1;
						}
					}
				}
			}
		}
		i=bomblist[i].next;
	} while (i>=0);
	
	return ret;
}

/*
   Control and draw the drones
*/

int HandleDrones(int b)
{
	int i,x,y,pdir,pcnt,newdir;
	int dirX[]={-4, 4, 4,-4};
	int dirY[]={-4,-4, 4, 4};
	
	for(i=0;i<numDrones;i++)
	{
		if (drones[i].dir==-1)
			continue;
		if (!drones[i].phase)
		{
			pdir=pcnt=0;
			if (tpos[drones[i].x-1][drones[i].y].type!=WATER && drones[i].x>0)
			{
				pdir|=1;
				pcnt++;
			}	
			if (tpos[drones[i].x][drones[i].y-1].type!=WATER && drones[i].y>0)
			{
				pdir|=2;
				pcnt++;
			}	
			if (tpos[drones[i].x+1][drones[i].y].type!=WATER && drones[i].x<field_width-1)
			{
				pdir|=4;
				pcnt++;
			}	
			if (tpos[drones[i].x][drones[i].y+1].type!=WATER && drones[i].y<field_width-1)
			{
				pdir|=8;
				pcnt++;
			}
			if (!pdir)
			{
				drones[i].dir=-1;
				continue;
			}
			
			if (!(pdir&(1<<drones[i].dir)) || pcnt>2)
			{
				if (pcnt>1)
					pdir&=~(1<<(drones[i].dir^2));

				for(;;)
				{
					newdir=rand()%4;
					if (pdir&(1<<newdir))
					{
						drones[i].dir=newdir;
						break;
					}
				}
			}
		}	
		x=tpos[drones[i].x][drones[i].y].x+dirX[drones[i].dir]*drones[i].phase;
		y=tpos[drones[i].x][drones[i].y].y+(dirY[drones[i].dir]*drones[i].phase*10)/16;
		TileCopy(b,DRONE_0+drones[i].phase/2,x,y-4,40,1);
		drones[i].phase=(drones[i].phase+1)%8;
		if (!drones[i].phase)
		{
			if (!dropShield && CheckDrop(drones[i].x,drones[i].y))
			{
				PlayWAV((void *)ScreamSound, -1, &Pan, 0x20000000, 0x10000000 );
				return -1;
			}
			switch(drones[i].dir)
			{
				case 0:
					drones[i].x--;
				break;	
				case 1:
					drones[i].y--;
				break;	
				case 2:
					drones[i].x++;
				break;	
				case 3:
					drones[i].y++;
				break;	
			}
			if (CheckSwitch(drones[i].x,drones[i].y)<0)
				return -2;
			if (!dropShield && CheckDrop(drones[i].x,drones[i].y))
			{
				PlayWAV((void *)ScreamSound, -1, &Pan, 0x20000000, 0x10000000 );
				return -1;
			}
		}	
	}
	return 0;	
}

/*
  Draw score, bonus, time, lives and level
*/

void RenderInfo(int drawBuffer,int what)
{
	char temp[256];
	
	if (score>99999) score=99999;
	
	if (what&1)
	{
		sprintf(temp,"%05d",score);
		PrintBig(drawBuffer,23,184,temp);
		PrintText(drawBuffer,23,208,"score");
	}
	if (what&2)
	{
		sprintf(temp,"%03d",bonus);
		PrintBig(drawBuffer,160,184,temp);
		PrintText(drawBuffer,160,208,"bonus");
	}
	if (what&4)
	{
//		sprintf(temp,"%04d",fps);
		sprintf(temp,"%04d",timer);
		PrintBig(drawBuffer,260,184,temp);
		PrintText(drawBuffer,260,208,"time");
	}
	if ((what&8) && lives>0)
	{
		sprintf(temp,"lives %d",lives-1);
		PrintText(drawBuffer,23,8,temp);
	}
	if (what&16)
	{
		sprintf(temp,"level %02d",level+1);
		PrintText(drawBuffer,240,8,temp);
	}
	if (dropSpinach)
	{
		sprintf(temp,"%02d",dropSpinach/10);
		PrintText(drawBuffer,170,8,temp);
	}
		
}

/*
    Level end, bonus calculation
*/

int LevelEnd(int drawBuffer)
{
	int joystick,amt;
	
	while(bonus>0)
	{
		joystick = _Controller[1].buttons |_Controller[0].buttons;                                                                               
		
		amt=(bonus>=5)?5:bonus;	

		if (joystick & JOY_A)
			amt=bonus;
			
		score+=amt;
		bonus-=amt;

		RenderWater(drawBuffer);
		RenderInfo(drawBuffer,31);
		DisplayTGA(drawBuffer,LevelCompTGA);
		
		drawBuffer=VideoSync(drawBuffer);
	}
    while(timer>0)
	{
		joystick = _Controller[1].buttons |_Controller[0].buttons;                                                                               
			
		amt=(timer>=20)?20:timer;	

		if (joystick & JOY_A)
			amt=timer;

		score+=amt;
		timer-=amt;

		RenderWater(drawBuffer);
		RenderInfo(drawBuffer,31);
		DisplayTGA(drawBuffer,LevelCompTGA);
		
		drawBuffer=VideoSync(drawBuffer);
	}	
	FadeOutAudio();
	return drawBuffer;
}

/*
   Calculate score control ID
*/


int ControlID(int value)
{
	int s,i,cc,m,ret;
		
	s=value;
	cc=0;
	i=10000;
	m=9;
	while(i)
	{
		cc+=(s/i)*m;
		s%=i;
		m--;
		i/=10;
	}
	ret=(((cc+99)/100)*100)-cc;
	return ret;
}

/*
   Game complete screen
*/

int GameComplete(int drawBuffer)
{
	char temp[256];
	
	sprintf(temp,"SCORE ID: %02d",ControlID(score));
	FadeOutAudio();
	RenderWater(drawBuffer);
	RenderInfo(drawBuffer,31);
	PrintText(drawBuffer,-1,120,"GAME COMPLETE!");
	PrintText(drawBuffer,-1,140,temp);
	PrintText(drawBuffer,-1,80,"BOMB VERSION '" BOMB_VERSION_STR "'");
	drawBuffer=VideoSync(drawBuffer);
	WaitForJoystick(JOY_A);
	RenderWater(drawBuffer);
	RenderInfo(drawBuffer,31);
	drawBuffer=VideoSync(drawBuffer);
	return drawBuffer;
}

/*
   Game-Over screen
*/

int GameOver(int drawBuffer)
{
	char temp[256];
	
	sprintf(temp,"SCORE ID: %02d",ControlID(score));
	FadeOutAudio();
	RenderWater(drawBuffer);
	RenderInfo(drawBuffer,31);
	DisplayTGA(drawBuffer,GameOverTGA);
	PrintText(drawBuffer,-1,140,temp);
	PrintText(drawBuffer,-1,80,"BOMB VERSION '" BOMB_VERSION_STR "'");
	drawBuffer=VideoSync(drawBuffer);
	WaitForJoystick(JOY_A);
	RenderWater(drawBuffer);
	RenderInfo(drawBuffer,31);
	drawBuffer=VideoSync(drawBuffer);
	return drawBuffer;
}

/*
   Timeout screen
*/

int TimeOut(int drawBuffer)
{
	FadeOutAudio();
	RenderWater(drawBuffer);
	RenderInfo(drawBuffer,31);
	PrintText(drawBuffer,-1,120,"TIME OUT");
	drawBuffer=VideoSync(drawBuffer);
	WaitForJoystick(JOY_A);
	RenderWater(drawBuffer);
	RenderInfo(drawBuffer,31);
	drawBuffer=VideoSync(drawBuffer);
	return drawBuffer;
}
 

/*
    Game entry point
*/


int main()
{
    int i,drawBuffer,fl;
	int edge,joystick;
	static int old_joystick=0;
	int paused=0;
	VidDisplay display;
	VidChannel mainchannel;
#ifdef OVERL_DEBUG
	VidChannel overlchannel;
#endif
	int analogX,analogY;

#ifdef OVERL_DEBUG
	if ((_Controller[1].buttons |_Controller[0].buttons) & JOY_UP)
		debug=1;
	else
		debug=0;                                                                             
#endif
	
	/* Init sound */
	
	InitSound();
	  
    Pan.PCMPanLR = 0;
    Pan.PCMPanFB = 0;
    Pan.PCMPanUD = 0;

    /* Get some local memory for DMA */

	localScratch=_MemLocalScratch(&localSize);
	
    /* Setup video */

    for(i=0;i<NUM_BUFFERS;i++)
	{
	    FrameBuffer[i]=_MemAlloc(SCREEN_WIDTH*SCREEN_HEIGHT*4,512,kMemSDRAM);
		myFillRect(FrameBuffer[i], DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, C32bppTo16bbp(kBlack));
	}
	
#ifdef OVERL_DEBUG
	if (debug)
		OverlayBuffer=_MemAlloc(SCREEN_WIDTH*SCREEN_HEIGHT*2,512,kMemSDRAM);
#endif
	
 	display.dispwidth = -1;
	display.dispheight = -1;
	display.bordcolor=kBlack;
	display.progressive=0;
	for(i=0;i<6;i++)
		display.reserved[i]=0;
	
	mainchannel.dmaflags=DMAFLAGS;
	mainchannel.base=FrameBuffer[0];
	mainchannel.dest_xoff = -1;
	mainchannel.dest_yoff = -1;
	mainchannel.dest_width = 720;
	mainchannel.dest_height = 480;
	mainchannel.src_xoff = 0;
	mainchannel.src_yoff = 0;
	mainchannel.src_width = SCREEN_WIDTH;
	mainchannel.src_height = SCREEN_HEIGHT;
	mainchannel.clut_select=0;
	mainchannel.alpha=0;
	mainchannel.vfilter=VID_VFILTER_2TAP;
	mainchannel.hfilter=VID_HFILTER_4TAP;
	for(i=0;i<5;i++)
		mainchannel.reserved[i]=0;

#ifdef OVERL_DEBUG
	if (debug)
	{
		overlchannel.dmaflags=OVERDMAFLAGS;
		overlchannel.base=OverlayBuffer;
		overlchannel.dest_xoff = -1;
		overlchannel.dest_yoff = -1;
		overlchannel.dest_width = 720;
		overlchannel.dest_height = 480;
		overlchannel.src_xoff = 0;
		overlchannel.src_yoff = 0;
		overlchannel.src_width = SCREEN_WIDTH;
		overlchannel.src_height = SCREEN_HEIGHT;
		overlchannel.clut_select=0;
		overlchannel.alpha=0;
		overlchannel.vfilter=0;
		overlchannel.hfilter=0;
		for(i=0;i<5;i++)
			overlchannel.reserved[i]=0;

		InitTerminalX(0,(int)OverlayBuffer,SCREEN_WIDTH, SCREEN_HEIGHT, OVERDMAFLAGS, 0x00);
		_VidConfig ( &display, &mainchannel, &overlchannel, 0L);
    }
	else
#endif
	_VidConfig ( &display, &mainchannel, 0, 0L); 
	
	/* show 'loading...' screen */
	
    DisplayTGA(0,LoadingTGA);
	
	srand(_VidSync(-1));
	vidOn=1;

	/* Initialize streaming audio */
	
	InitStreamingAudio();

	/* Game restart point */
	
RestartGame:
	drawBuffer=0;
	_VidChangeBase(VID_CHANNEL_MAIN,DMAFLAGS,FrameBuffer[drawBuffer]);

	for(;;)
	{
		myFillRect(FrameBuffer[0], DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, C32bppTo16bbp(kBlack));
		myFillRect(FrameBuffer[1], DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, C32bppTo16bbp(kBlack));
	    DisplayTGA(0,StartScreenTGA);
		_VidChangeBase(VID_CHANNEL_MAIN,DMAFLAGS,FrameBuffer[0]);
		i=WaitForJoystick(JOY_START|JOY_A);
		if (i&JOY_A)
		{
    		 DisplayTGA(1,Instr1TGA);
			_VidChangeBase(VID_CHANNEL_MAIN,DMAFLAGS,FrameBuffer[1]);
			 WaitForJoystick(JOY_START|JOY_A);
   		     DisplayTGA(1,Instr2TGA);
			 WaitForJoystick(JOY_START|JOY_A);
		}
		else
			break;
	}
	score = 0;
	level = 0;
	lives = 3;		

	/* New level start point */

NextLevel:

	
	field_width = levels[level].size;
	field = levels[level].field;
	switches = levels[level].switches;
	bonus = field_width*field_width*5;
	bombTime = levels[level].bomb_time;	

	/* Restart level start point */

RestartLevel:

	StartSong(level);
	FadeInAudio();

	timer = levels[level].time;

	lastTileType=TILE;
	dropDir=2;
	dropPhase=dropRotate=dropAnimation=dropIdle=dropShield=dropSpinach=bombMove=0;
	dropTimer=0;
	field_x=(VW-64*field_width)/2+VX;
	field_y=(VH-40*field_width)/2+VY;
	moveField=0;
	ResetPlayField();
    FieldFromASCII(field_width);
	dropX=dropStartX;
	dropY=dropStartY;

    /* main game loop */
	drawBuffer=1;
	for(;;)
	{
		joystick = _Controller[1].buttons |_Controller[0].buttons;                                                                               
		edge = (joystick ^ old_joystick) & joystick; 
		old_joystick = joystick;
		
		if (score>=newlifeThres)
		{
			lives++;
			newlifeThres+=1000;
   			PlayWAV((void *)SpinachSound, -1, &Pan, 0x20000000, 0x00000000 );
		}
		
		/* Load more audio data, if necessary */
		HandleStreamingAudio();
		
		analogX=(_Controller[1].d1.xAxis/16)*32;
		analogY=(_Controller[1].d2.yAxis/16)*32;
		
		/* Render background and the playfield */
		RenderWater(drawBuffer);
		RenderField(drawBuffer,field_width,field_x+analogX,field_y+analogY);
		
		if (!paused)
		{
			if (edge&JOY_START)
				paused=1;

			/* Render info */
			RenderInfo(drawBuffer,31);
			if ((dropAnimation&0xffff)!=DROP_DROWNING)
			{
				/* Move drones */
				fl=HandleDrones(drawBuffer);
				if (fl<0)
				{
					if (fl==-2)
						dropAnimation=DROP_START_DROWNING;   
					else
						break;
				}
			}
			
			/* Handle timer */

			timer--;
			if (!timer)
			{
				drawBuffer=TimeOut(drawBuffer);
				break;
			}
			
			/* Handle shield and spinach timer */
			if (dropShield) dropShield--;
			if (dropSpinach) dropSpinach--;
			
			/* Handle Drop */
			fl=ControlDrop(drawBuffer);
			if (fl==-1)
				break;


			/* Check if level has completed */
			if (bombs<=0 && !dropAnimation && !dropPhase && bombListStart<0)
			{
				drawBuffer=LevelEnd(drawBuffer);
				level++;
				if (level==NUM_LEVELS)
				{
					level--;
					drawBuffer=GameComplete(drawBuffer);
					goto RestartGame;
				}
				goto NextLevel;
			}
			
			
			/* Handle camera */
			MovePlayfield(fl);
		}
		else
		{
			/* we are paused ... */
			DisplayTGA(drawBuffer,PausedTGA);
			if (edge&JOY_START)
				paused=0;
		}
		
		/* Handle bomns and explosions */
		if (HandleBombs(drawBuffer)<0)
			break;

		if (edge&JOY_L)
	  		SaveRGBFrame(drawBuffer);

		/* Synchronize with VBL and switch to next framebuffer */		
		drawBuffer=VideoSync(drawBuffer);
	}
	
	/* we are dead ... */
	lives--;
	if (lives)
	{
		FadeOutAudio();
		_VidSync(100);
		goto RestartLevel;
	}
	else
	{
		GameOver(drawBuffer);
		goto RestartGame;
	}
}
