/*
    BOMB, an open-platform NUON game
	
	Programming by Andreas Binner
	
	Graphics module
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <nuon/sprite.h>
#include <nuon/bios.h>
#include <nuon/dma.h>
#include <nuon/video.h>
#include <nuon/joystick.h>
#include <nuon/mediaio.h>
#include <nuon/nise.h>
#include <nuon/termemu.h>

#include "def.h"
#include "proto.h"

extern long Tiles[];
extern long BombFont[];
extern unsigned long *localScratch;
extern void *FrameBuffer[];

typedef struct
{
	int x,y;
} TILESRC;

/* Tile positions in the tiles.ppm image */

TILESRC tiletab[]={
{0  ,0},      // 0
{64 ,0},
{0  ,40},
{64 ,40},
{0  ,80},
{0  ,120 },
{64 ,80  },
{128 ,0  },
{128 ,48 },
{128 ,96 },   // 10
{128 ,144},
{192 ,0  },
{192 ,48 },
{192 ,96 },
{192 ,144},
{256 ,0  },
{256 ,48 },
{256 ,96 },
{256 ,144},
{320 ,0  },   // 20
{320 ,48 },
{320 ,96 },
{320 ,144},
{128 ,320},
{0  ,160 },
{0  ,200 },
{0  ,240 },
{0  ,280 },
{64 ,280 },
{64 ,240 },
{64 ,200 },   // 30
{64 ,120 },
{64 ,160 },
{128 ,240},
{128 ,280},
{192 ,240},
{192 ,280},
{256 ,240},
{256 ,280},
{320 ,240},
{320 ,280},   // 40
{256 ,320},
{320 ,320},
{192 ,320},
{ 64 ,320}
};

/* Clip to unsigned 8-bit */

int Clip(int v)
{
    if (v<0) return 0;
    if (v>255) return 255;
    return v;
}

/* For screen shots. Works only on development systems! */

void SaveRGBFrame(int b)
{
    int x, y, i;
 	long *save = (long *)0x80800000;
    int crv,cbu,cgu,cgv,Y,Cr,Cb,R,G,B;
	
	crv = 117504;
    cbu = 138453;
    cgu = 13954;
    cgv = 34903;
	
    for (y = 0; y < SCREEN_HEIGHT; y ++) {
	    for (x = 0; x < SCREEN_WIDTH; x += 8) {
			_DMABiLinear(DMAFLAGS|(1<<13),FrameBuffer[b],(8<<16)|x,(1<<16)|y,localScratch);
            for(i=0;i<8;i++)
			{
//				kprintf("%08lx: ",localScratch[i]);
				Y = 76309 * (((localScratch[i]>>26)<<2) - 16);
            	Cr= (((localScratch[i]>>21)&0x1f)<<3) - 128;
            	Cb= (((localScratch[i]>>16)&0x1f)<<3) - 128;

//				kprintf("%d/%d/%d -> ",Y,Cr,Cb);

            	R = Clip((Y + crv*Cr + 32768)>>16);
            	G = Clip((Y - cgu*Cb - cgv*Cr + 32768)>>16);
            	B = Clip((Y + cbu*Cb + 32786)>>16);

//				kprintf("%d/%d/%d: ",R,G,B);
				
				localScratch[i] = (R<<24) + (G<<16) + (B<<8);

//				kprintf("%08lx\n",localScratch[i]);
			}
//			WaitForJoystick(JOY_A);
			_DMALinear(8<<16,&save[x+y*SCREEN_WIDTH],localScratch);

        }
    }
}

/* Display a centered TGA */

int DisplayTGA(int b,unsigned char *tga)
{
	long w,h;
	
	GetTGASize(tga, &w, &h);
	return ReadTGA(tga, DMAFLAGS|(4<<1), FrameBuffer[b] ,(SCREEN_WIDTH-w)/2, (SCREEN_HEIGHT-h)/2, 0, 0xff00ff00);

}

/*
    Simple routine to draw a filled rectangle (via DMA)
*/

void myFillRect(char *screen, int dmaflags, int initx, int inity, int wide, int high, int color)
{
    int x, y;
    int lastx, lasty;
 
    lastx = initx + wide;
    lasty = inity + high;
 
    for (x = initx; x < lastx; x += 8) {
                for (y = inity; y < lasty; y += 8) {
                _raw_plotpixel(dmaflags, (void *)screen, (8<<16)|x, (8<<16)|y, color);
                }
    }
}

/*
    Draws one character from the font.ppm 
*/

void DigitCopy(int b,char c,int x,int y)
{
	int i,sx,sy;
	
	if (c>='0' && c<='9')
	{
		sx=CHAR_WIDTH*(int)(c-'0');
		sy=0;
	}
	else
	if (c>='A' && c<='Z')
	{
		sx=CHAR_WIDTH*(int)(c-'A');
		sy=16;
	}
	else
	if (c>='a' && c<='z')
	{
		sx=CHAR_WIDTH*(int)(c-'a');
		sy=16;
	}
	else
	if (c>='!' && c<='/')
	{
		sx=CHAR_WIDTH*(int)(c-'!');
		sy=32;
	}
	else
		return;
		
	for(i=sy;i<sy+16;i++)
	{
		_DMALinear(CHAR_WIDTH<<16|DMA_READ_BIT,&BombFont[i*FONT_WIDTH+sx],localScratch);
		_DMABiLinear(DMAFLAGS|(4<<1),FrameBuffer[b],(CHAR_WIDTH<<16)|x,(1<<16)|y,localScratch);
		y++;
	}
}

/*
   Drows one big digit from the tiles.ppm image
*/

void DigitCopyX(int b,int n,int x,int y,int w)
{
	int i;
	
	for(i=205;i<228;i++)
	{
		_DMALinear(w<<16|DMA_READ_BIT,&Tiles[i*SRC_WIDTH+128+n*w],localScratch);
		_DMABiLinear(DMAFLAGS|(4<<1),FrameBuffer[b],(w<<16)|x,(1<<16)|y,localScratch);
		y++;
	}
}

/*
    Print a text string (big digits only!)
*/

void PrintBig(int b,int x,int y,char *str)
{
	while(*str)
	{
		if (*str>='0' && *str<='9')
			DigitCopyX(b,(int)*str-'0',x,y,20);
		x+=18;
		str++;
	}
}

/*
    Print a text string
*/

void PrintText(int b,int x,int y,char *str)
{
	if (y<0)
		y=(SCREEN_HEIGHT-20)/2;
	if (x<0)
		x=(SCREEN_WIDTH-strlen(str)*(CHAR_WIDTH-3))/2;
				
	while(*str)
	{
		DigitCopy(b,*str,x,y);
		x+=CHAR_WIDTH-3;
		str++;
	}
}

/*
   Draws one tile (fast version, no clipping!)
*/

void FastTileCopy(int b,int type,int dx,int dy,int h)
{
	int tx,ty,y;
	long *t;
		
	tx=tiletab[type].x;
	ty=tiletab[type].y;	
	t=&Tiles[ty*SRC_WIDTH+tx];		
	for(y=ty;y<ty+h;y++)
	{
		_DMALinear(64<<16|DMA_READ_BIT,t,localScratch);
		_DMABiLinear(DMAFLAGS,FrameBuffer[b],(64<<16)|dx,(1<<16)|dy,localScratch);
		dy++;
		t+=SRC_WIDTH;
	}

}

/*
   Draws one tile (version does clipping, teleport and shield effect)
*/

void TileCopy(int b,int type,int dx,int dy,int h,int flags)
{
	int tx,ty,y,fl,i;
	long *t;
	
	if (flags&1)
		fl=(4<<1);
	else
		fl=0;
		
	tx=tiletab[type].x;
	ty=tiletab[type].y;	
	t=&Tiles[ty*SRC_WIDTH+tx];		
			
	for(y=ty;y<ty+h;y++)
	{
		_DMALinear(64<<16|DMA_READ_BIT,t,localScratch);
		if (flags&0x2)
			for(i=0;i<64;i++)
				localScratch[i] = (rand()&0x1)?0x00000000:localScratch[i];
		if (flags&0x4)
			for(i=0;i<64;i++)
				localScratch[i] = localScratch[i]&0xfcccffff;
		if (dy>=0 && dy<SCREEN_HEIGHT)
		{
			if (dx>=0 && dx<=SCREEN_WIDTH-64)
				_DMABiLinear(DMAFLAGS|fl,FrameBuffer[b],(64<<16)|dx,(1<<16)|dy,localScratch);
			else
			{
				if (dx>-64 && dx<SCREEN_WIDTH)
				{
					if (dx<0)
						_DMABiLinear(DMAFLAGS|fl,FrameBuffer[b],((64+dx)<<16)|0,(1<<16)|dy,&localScratch[-dx]);
					else
						_DMABiLinear(DMAFLAGS|fl,FrameBuffer[b],((SCREEN_WIDTH-dx)<<16)|dx,(1<<16)|dy,localScratch);
				}
			}
		}
		dy++;
		t+=SRC_WIDTH;
	}

}
