#define NUM_LEVELS 14
/* Change below name to something new and unique when changing level design !!! */
#define BOMB_VERSION_STR "NEWYEAR"

typedef struct
{
	int time;          /* Timer start value */
	int size;          /* Width of the playfield (which has to be square!) */
	int bomb_time;     /* Time between bomb activation and explosion */
	char *field;       /* playfield definition */
	char *switches;    /* Switchable fields definitions*/
} LEVEL;


/*
    The playfield is defined as one 'size*size' byte long string:
	
	" "  Water
	"#"  Regular tile
	"*"  Ice
	"x"  Broken tile
	"."  Small bomb
	"o"  Medium bomb
	"O"  Big Bomb
	">"  Small mine
	"<"  Big Mine
	"~"  Shield
	"="  Unbreakable tile
	"$"  Drop start position
	"?"  Drone start position (max. 4 drones!)
	"S"  Spinach
	
	"a"-"f"  Switch 0-5
	"A"-"F"  Switchable tile activated by switch 0-5
	
	"i"-"n"  Teleporter 0-5
	"I"-"N"  Teleporter destination 0-5
	
	Switchable tiles are defined in one extra string. For each switch
	three bytes are used:
	
	Byte 0:  Switch "a"-"f"
	Byte 1: "ON" tile definition
	Byte 2: "OFF" tile definition
	
	Example:
	
	"a#*" -> Switchable tile for switch 0 ("A") is a regular tile ("#") in
	         "ON" state and ice ("*") in "OFF" state.
		
*/

LEVEL levels[NUM_LEVELS] =
{
	{
		500,
		4,
		30,
		" .$ "
		" x.."
		"..x "
		" #.#",

		"" 
	},
	{
		700,
		8,
		30,
		"$*.*.*#*"
		"********"
		" **.***."
		".*****.*"
		"*.**.***"
		".******."
		"*.*.****"
		"********",

		"" 
	},
	{
		500,
		7,
		30,
		"   *   "
    	" *$o.# "
		" x *   "
		" N.*.n "
		"   * o "
		" ..<.o "
		"   *   ",

		""
	},
	{
		500,
		8,
		40,
		"fno#o#o#"
		"O      o"
		"<  EF  #"
		"x$ .*  o"
		"o   *=##"
		"#   *  o"
		"o   M  #"
		"moeoNo#o",

		"e= f# "
	},
	{
		700,
		6,
		30,
		".#***o"
		"  o  #"
		". .  o"
		"$####."
		"     #"
		" #Oo.#",

		""
	},
	{
		900,
		6,
		30,
		"N>K*l."
		" .#*.#"
		".# M**"
		"*.mk= "
		"*o Soo"
		"*x$n.L",

		""
	},
	{
		1000,
		8,
		30,
		".##<#F#."
		"# D #  #"
		"#e*#f  *"
		"C * x$ *"
		"O *  x *"
		"# x  d.E"
		"#*## #  "
		"   >c#  ",

		"c =dx ex fx " 
	},
	{
		800,
		7,
		65,
		"####>##"
		"#  $  x"
		"e >#=E#"
		"    # #"
		"##o#F f"
		"#      "
		"i#o#I..",

		"e= fS#" 
	},
	{
		600,
		8,
		30,
		" I  xxJx"
		"OOO ..b."
		"O$Oxxxxo"
		"CAB ..d."
		" #K#   D"
		"   .  #a"
		"jik   .#"
		"xxxc***#",

		"aO.b.ocOod# "

	},
	{
		800,
		8,
		30,
		"xx  xbxx"
		"$#a I##o"
		"xxx *  ."
		"  A *  x"
		"..o *c.#"
		"ijk    C"
		".K. Jo.."
		"...    B",

		"ax b= cx "

	},
	{
		900,
		8,
		65,
		" ##n*#o "
		" .# *## "
		"A***o#o#"
		" b     #"
		"$#FEDCo#"
		" N      "
		"       a"
		" ?fedc#B",

		"a# b# c #d #e #f# "
	},
	{
		800,
		8,
		30,
		"***   **"
		"$aA.# .~"
		"#. #bBo<"
		"  ..# .."
		"#..#. #o"
		"#..#. #o"
		".#..? ##"
		"..    **",

		"ax bx "

	},
	{
		500,
		8,
		30,
		"?##>#>C>"
		"F D  # #"
		"d *#*. f"
		". *.**EN"
		"  *..# c"
		"#####~ #"
		"x      #"
		"$>n.#e#?",

		"c# d ~e# f# "
	
	},
	{
		3200,
		16,
		30,
		".      ###   ###"
		"#~#xx###e### # ."
		" #         # # #"
		" # . ##### # # #"
		"## # #   # # . #"
		"#d # .## # #   #"
		"#  #   # #f# #.#"
		"#  # # $   # #  "
		"#x####   # E #.#"
		"# # #F#~#### # #"
		"o . # #    # # #"
		"### # # ###c ## "
		"    # . #     # "
		"##### # # ####D "
		"#     # # #     "
		"#########C#####?",

		"c# d*#e# f# "
	
	}
};
